------------------------------------------------------------------------------
--				Highlight code for plenty of conditions						--
--	supports Buffs, Debuffs, Equiped items, out of mana, no range,			--
--	not usable, currently casting, currently channeling and autoattacking	--
------------------------------------------------------------------------------
local Highlighting =  FlexBar2:NewModule("Highlighting");
-- If this var is incremented, SetDefaults() will be called for this module thus resetting all its settings. Though there are no settings for this module, it has to be declared
Highlighting.Version = 1;
-- Get a local version of SpecialEvents-Aura
local Aura = Rock("LibSpecialEvents-Aura-3.0");

-- Note: Both Inner & Outer Priority will be configurable in the future
-- Define priority for inner checks (mana, range & usable)
local InnerPriority = {"NoRange", "NoMana", "Unusable"};
-- Define priority for outer checks (channeling, casting, autoattacking, buffed, debuffed, equipped)
local OuterPriority = {"Channeling", "Casting","AutoCasting", "ItemBuff", "Buff", "Debuff", "Equipped"};
local RangeGuess;

-- Register all colors with paintchips, scheduled to be configurable in the future
function Highlighting:OnInitialize()
	-- Keep a hash table of RGB values for highlights
	local ColorList = {};
	self.ColorList = ColorList;
	ColorList.Buff = {0, 255, 0};
	ColorList.ItemBuff = {186, 0, 211};
	ColorList.Debuff = {255 , 0, 0};
	ColorList.Equipped = {255 , 255, 0};
	ColorList.NoMana = {0 , 0, 150};
	ColorList.NoRange = {150 , 0, 0};
	ColorList.Unusable = {100 , 100, 100};
	ColorList.Casting = {255, 255, 255};
	ColorList.Channeling = {152, 245, 255};
	ColorList.AutoCasting = {238, 118, 0};
end

-- Declare the button mixin
Highlighting.ButtonMixin = {};

local ButtonCheckUsable = function(Button, Event, NameSpace, Source, UnitId) if(UnitId == "player") then self:CheckUsable(); end end;
local ButtonCheckEquipped = function(Button, Event, NameSpace, Source, UnitId) if(UnitId == "player") then self:CheckEquipped(); end end;
local ButtonCheckCasting = function(Button, Event, NameSpace, Source, UnitId) if(UnitId == "player") then self:CheckCasting(); end end;
local ButtonCheckChanneling = function(Button, Event, NameSpace, Source, UnitId) if(UnitId == "player") then self:CheckChanneling(); end end;
local ButtonCheckBuff = function(self, Event, NameSpace, Dispatcher, Button) if(Button == self) then self:CheckBuff(); end end;
local ButtonCheckItemBuff = function(self, Event, NameSpace, Dispatcher, Button) if(Button == self) then self:CheckItemBuff(); end end;
local ButtonCheckDebuff = function(self, Event, NameSpace, Dispatcher, Button) if(Button == self) then self:CheckDebuff(); end end;
local ButtonCheckEquipped = function(self, Event, NameSpace, Dispatcher, Button) if(Button == self) then self:CheckEquipped(); end end;
local ButtonCheckUsable = function(self, Event, NameSpace, Dispatcher, Button) if(Button == self) then self:CheckUsable(); end end;
local InitHighlights = function(self)
	self:CheckEquipped();
	self:CheckBuff();
	self:CheckDebuff();
	self:CheckUsable();
	self:CheckItemBuff();
end

function Highlighting.ButtonMixin:Activate()
	-- Variables to store what checks are currently active
	self.InnerCheck = {};
	self.OuterCheck = {};
	-- Update the outer highlight when clicked, as clicking might check/uncheck it
	self:RegisterScript("PostClick", "UpdateOuterHighlight");
	-- Update range highlight every 0.3 sec with FB2_UPDATE
	self:AddEventListener("FlexBar2_EventManager", "UPDATE", "CheckRange");
	-- Update usability on mana, rage & energy updates (this includes mana)
	self:AddEventListener("UNIT_MANA", "CheckUsable");
	self:AddEventListener("UNIT_RAGE", "checkUsable");
	self:AddEventListener("UNIT_ENERGY", "CheckUsable");
	-- Update equiped items when inventory changes
	self:AddEventListener("UNIT_INVENTORY_CHANGED", "CheckEquipped");
	-- Update casting on spellcasting start & stop
	-- Currently buggy so has to be done OnUpdate :'(
	self:AddEventListener("FlexBar2_EventManager", "UPDATE", "CheckCasting");

	--self:AddEventListener("UNIT_SPELLCAST_START", ButtonCheckCasting);
	--self:AddEventListener("UNIT_SPELLCAST_STOP", ButtonCheckCasting);
	--]]
	-- Update channeling on channeling start & stop
	-- Currently buggy so has to be done OnUpdate :(
	self:AddEventListener("FlexBar2_EventManager", "UPDATE", "CheckChanneling");

	--self:AddEventListener("UNIT_SPELLCAST_CHANNEL_START", ButtonCheckChanneling);
	--self:AddEventListener("UNIT_SPELLCAST_CHANNEL_STOP", ButtonCheckChanneling);
	-- Update autoattacking (or wanding, shooting, whatever) with StartAutoRepeat and EnterCombat events
	self:AddEventListener("START_AUTOREPEAT_SPELL", "CheckStartAutoCasting");
	self:AddEventListener("STOP_AUTOREPEAT_SPELL", "CheckStopAutoCasting");
	self:AddEventListener("PLAYER_ENTER_COMBAT", "CheckStartAutoCasting");
	self:AddEventListener("PLAYER_LEAVE_COMBAT", "CheckStopAutoCasting");
	-- Update buffs when SpecialEvent triggers on player buffs
	self:AddEventListener("LibSpecialEvents-Aura-3.0", "PlayerBuffGained", "CheckBuff");
	self:AddEventListener("LibSpecialEvents-Aura-3.0", "PlayerBuffLost", "CheckBuff");
	-- Update usability when the player gains debuffs (hint: forbearance)
	self:AddEventListener("LibSpecialEvents-Aura-3.0", "PlayerDebuffGained", "CheckUsable");
	self:AddEventListener("LibSpecialEvents-Aura-3.0", "PlayerDebuffLost", "CheckUsable");
	-- Update debuffs when the target gains a debuff
	self:AddEventListener("LibSpecialEvents-Aura-3.0", "UnitDebuffGained", "CheckDebuff");
	self:AddEventListener("LibSpecialEvents-Aura-3.0", "UnitDebuffLost", "CheckDebuff");
	self:AddEventListener("LibSpecialEvents-Aura-3.0", "AuraTargetChanged", "CheckDebuff");
	-- Check for item buffs
	self:AddEventListener("LibSpecialEvents-Aura-3.0", "PlayerItemBuffGained", "CheckItemBuff");
	self:AddEventListener("LibSpecialEvents-Aura-3.0", "PlayerItemBuffLost", "CheckItemBuff");
	-- Update everything if the action changed for this button
	self:AddEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckBuff);
	self:AddEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckDebuff);
	self:AddEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckChanneling);
	self:AddEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckCasting);
	self:AddEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckEquipped);
	self:AddEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckUsable);
	self:AddEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckItemBuff);

	-- Highlight upon loading too
	if(Rock("LibRockEvent-1.0"):IsFullyInitialized()) then
		InitHighlights(self);
	else
		self:AddEventListener("LibRockEvent-1.0", "FullyInitialized", InitHighlights);
	end
end

function Highlighting.ButtonMixin:Deactivate()
	self:RemoveEventListener("FlexBar2_EventManager", "UPDATE", "CheckRange");
	self:RemoveEventListener("UNIT_INVENTORY_CHANGED", ButtonCheckEquipped);
	self:RemoveEventListener("UNIT_MANA", "CheckUsable");
	self:RemoveEventListener("UNIT_RAGE", "CheckUsable");
	self:RemoveEventListener("UNIT_ENERGY", "CheckUsable");
	self:RemoveEventListener("FlexBar2_EventManager", "UPDATE", "CheckCasting");
	self:RemoveEventListener("FlexBar2_EventManager", "UPDATE", "CheckChanneling");
	self:RemoveEventListener("START_AUTOREPEAT_SPELL", "CheckStartAutoCasting");
	self:RemoveEventListener("STOP_AUTOREPEAT_SPELL", "CheckStopAutoCasting");
	self:RemoveEventListener("PLAYER_ENTER_COMBAT", "CheckStartAutoCasting");
	self:RemoveEventListener("PLAYER_LEAVE_COMBAT", "CheckStopAutoCasting");
	self:RemoveEventListener("LibSpecialEvents-Aura-3.0", "PlayerBuffGained", "CheckBuff");
	self:RemoveEventListener("LibSpecialEvents-Aura-3.0", "PlayerBuffLost", "CheckBuff");
	self:RemoveEventListener("LibSpecialEvents-Aura-3.0", "PlayerDebuffGained", "CheckUsable");
	self:RemoveEventListener("LibSpecialEvents-Aura-3.0", "PlayerDebuffLost", "CheckUsable");
	self:RemoveEventListener("LibSpecialEvents-Aura-3.0", "UnitDebuffGained", "CheckDebuff");
	self:RemoveEventListener("LibSpecialEvents-Aura-3.0", "UnitDebuffLost", "CheckDebuff");
	self:RemoveEventListener("LibSpecialEvents-Aura-3.0", "AuraTargetChanged", "CheckDebuff");
	self:RemoveEventListener("LibSpecialEvents-Aura-3.0", "PlayerItemBuffGained", "CheckItemBuff");
	self:RemoveEventListener("LibSpecialEvents-Aura-3.0", "PlayerItemBuffLost", "CheckItemBuff");
	self:RemoveEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckBuff);
	self:RemoveEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckDebuff);
	self:RemoveEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckChanneling);
	self:RemoveEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckCasting);
	self:RemoveEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckEquipped);
	self:RemoveEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckUsable);
	self:RemoveEventListener("FlexBar2_EventManager", "ActionChanged", ButtonCheckItemBuff);
end

-- Make sure the outer highlight is set correctly, called on postclick and when a highlight changes
function Highlighting.ButtonMixin:UpdateOuterHighlight()
	if(type(self.OuterCheck[1]) == "string") then
		self.Frame:SetChecked(true);
		self.Frame:GetCheckedTexture():SetVertexColor(unpack(Highlighting.ColorList[self.OuterCheck[1]]));
	else
		self.Frame:SetChecked(false);
	end
end

local function InnerSort(a, b)
	return FlexBar2:GetKeyByValue(InnerPriority, a) < FlexBar2:GetKeyByValue(InnerPriority, b);
end

local function OuterSort(a, b)
	FlexBar2:Print(a, b);
	return FlexBar2:GetKeyByValue(OuterPriority, a) < FlexBar2:GetKeyByValue(OuterPriority, b);
end

function Highlighting.ButtonMixin:SetInnerHighlight(CheckType, State)
	-- State is true, or an equivalent of true
	if(State) then
		-- Make sure this highlight is not already active
		if(not FlexBar2:ValueIsInTable(self.InnerCheck, CheckType)) then
			-- insert the highlight into the table
			table.insert(self.InnerCheck, CheckType);
			-- sort the table with the priorities in our priority tables declared at the top
			table.sort(self.InnerCheck, InnerSort);
		end
	-- State is false or an equivalent of false
	else
		-- Remove the highlight
		FlexBar2:RemoveByValue(self.InnerCheck, CheckType);
	end
	-- find out if there's still a highlight active after doing our stuff
	if(type(self.InnerCheck[1]) == "string") then
		-- If there's a highlight active, set it
		self.Frame.Icon:SetVertexColor(unpack(Highlighting.ColorList[self.InnerCheck[1]]));
	else
		-- If there's no highlight active, back to plain boring white
        self.Frame.Icon:SetVertexColor(1, 1, 1);
	end
end

function Highlighting.ButtonMixin:SetOuterHighlight(CheckType, State)
	-- State is true, or an equivalent of true
	if(State) then
		-- Make sure this highlight is not already active
		if(not FlexBar2:ValueIsInTable(self.OuterCheck, CheckType)) then
			-- insert the highlight into the table
			table.insert(self.OuterCheck, CheckType);
			-- sort the table with the priorities in our priority tables declared at the top
			table.sort(self.OuterCheck, OuterSort);
		end
	else
		FlexBar2:RemoveByValue(self.OuterCheck, CheckType)
	end
	-- Update the outer highlight
	self:UpdateOuterHighlight();
end

-- Check if the user is currently auto casting something, be it autoattack, autowand or autoshot
function Highlighting.ButtonMixin:CheckStartAutoCasting()
	local Spell = self:GetModifiedAttribute("spell"); 
	if(Spell) then
		self:SetOuterHighlight("AutoCasting", IsAutoRepeatSpell(Spell) or IsAttackSpell(Spell));
	else
		self:SetOuterHighlight("AutoCasting", false);
	end
end

-- Triggered when autocasting stops
function Highlighting.ButtonMixin:CheckStopAutoCasting()
	self:SetOuterHighlight("AutoCasting", false);
end

-- Grab a local copy
local GetSpellInfo = FlexBar2:GetModule("SpellCache").GetSpellInfo;
function Highlighting.ButtonMixin:CheckCasting()
	local CurrentSpellName, CurrentSpellRank = UnitCastingInfo("player");
	local CurrentSpell;
	-- Only continue if we're actualy casting stuff
	if(CurrentSpellName) then
		if(CurrentSpellRank) then
			CurrentSpell = CurrentSpellName .. "(" .. CurrentSpellRank .. ")";
		else
			CurrentSpell = CurrentSpellName;
		end
		-- Ok, the picture here is, sometimes we don't specify a rank, so we have to figure what rank is the highest one, do this will GetSpellName
		local ButtonSpell = self:GetModifiedAttribute("spell")
		if(ButtonSpell) then
			local ButtonSpellName, ButtonSpellRank = GetSpellName(GetSpellInfo(nil, ButtonSpell), BOOKTYPE_SPELL);
			if(ButtonSpellRank) then
				ButtonSpell = ButtonSpellName .. "(" .. ButtonSpellRank .. ")";
			else
				ButtonSpell = ButtonSpellName;
			end
			self:SetOuterHighlight("Casting", ButtonSpell == CurrentSpell);
		else
			self:SetOuterHighlight("Casting", false);
		end
	else
		self:SetOuterHighlight("Casting", false);
	end
end

-- Checks if the user is currently casting something
--function Highlighting.ButtonMixin:CheckCasting()
	-- the IsCurrentSpell function used here comes out in 2.3
	--[[
	local ActionType = self:GetModifiedAttribute("type");
	if(ActionType == "spell") then
		self:SetOuterHighlight("Casting", IsCurrentSpell(self:GetModifiedAttribute("spell")));
	elseif(ActionType == "item") then
		self:SetOuterHighlight("Casting", IsCurrentItem(self:GetModifiedAttribute("itemlink")));
	elseif(ActionType == "macro") then
		-- Not supported yet
		self:SetOuterHighlight("Casting", false);
	else
		self:SetOuterHighlight("Casting", false);
	end-]]
--end

-- Checks for channeled spells
function Highlighting.ButtonMixin:CheckChanneling()
	local CurrentSpellName, CurrentSpellRank = UnitChannelInfo("player");
	local CurrentSpell;
	-- Only continue if we're actualy casting stuff
	if(CurrentSpellName) then
		if(CurrentSpellRank) then
			CurrentSpell = CurrentSpellName .. "(" .. CurrentSpellRank .. ")";
		else
			CurrentSpell = CurrentSpellName;
		end
		-- Ok, the picture here is, sometimes we don't specify a rank, so we have to figure what rank is the highest one, do this will GetSpellName
		local ButtonSpell = self:GetModifiedAttribute("spell")
		if(ButtonSpell) then
			local ButtonSpellName, ButtonSpellRank = GetSpellName(GetSpellInfo(nil, ButtonSpell), BOOKTYPE_SPELL);
			if(ButtonSpellRank) then
				ButtonSpell = ButtonSpellName .. "(" .. ButtonSpellRank .. ")";
			else
				ButtonSpell = ButtonSpellName;
			end
			self:SetOuterHighlight("Channeling", ButtonSpell == CurrentSpell);
		else
			self:SetOuterHighlight("Channeling", false);
		end
	else
		self:SetOuterHighlight("Channeling", false);
	end
end

-- Checks for itembuffs on the player
function Highlighting.ButtonMixin:CheckItemBuff()
	local MainBuff, MainRank = Aura:GetPlayerMainHandItemBuff();
	local OffBuff, OffRank = Aura:GetPlayerOffHandItemBuff();
	if(not MainBuff and not OffBuff) then
		self:SetOuterHighlight("ItemBuff", false);
		return;
	end
	local ButtonType = self:GetModifiedAttribute("type");
	if(not ButtonType) then return; end
	if(ButtonType == "macro") then --[[ TODO: Add code here in 2.3]]  return; end
	local ButtonAction = self:GetModifiedAttribute(ButtonType);
	if((MainBuff or OffBuff) and ButtonAction) then
		if(ButtonType == "spell") then
			local ButtonSpell = ButtonAction;
			local ButtonSpellName, ButtonSpellRank = GetSpellName(GetSpellInfo(nil, ButtonSpell), BOOKTYPE_SPELL);
			if(ButtonSpellRank) then
				local ButtonSpell = ButtonSpellName .. "(" .. ButtonSpellRank .. ")";
				-- Try with the Weapon suffix (shammy self buffs)
				if(MainBuff and ButtonSpell == (MainBuff .. " Weapon" .. (MainRank and "(Rank " .. MainRank .. ")" or ""))) then
					self:SetOuterHighlight("ItemBuff", true);
					return;
				end
				if(OffBuff and ButtonSpell == (OffBuff .. " Weapon" .. (OffRank and "(Rank " .. OffRank .. ")" or ""))) then 
					self:SetOuterHighlight("ItemBuff", true);
					return;
				end
				-- Try with totem suffix (shammy totem buffs)
				if(MainBuff and ButtonSpell == (MainBuff .. " Totem" .. (MainRank and "(Rank " .. MainRank .. ")" or ""))) then
					self:SetOuterHighlight("ItemBuff", true);
					return;
				end

				if(OffBuff and ButtonSpell == (OffBuff .. " Totem" .. (OffRank and "(Rank " .. OffRank .. ")" or ""))) then 
					self:SetOuterHighlight("ItemBuff", true);
					return;
				end
				-- Fail
				self:SetOuterHighlight("ItemBuff", false);
				return;

			end
			elseif(ButtonType == "item") then
			local ItemName = GetItemInfo(ButtonAction);
			self:SetOuterHighlight("ItemBuff", ItemName == MainBuff or ItemName == OffBuff);
		end
	else
		self:SetOuterHighlight("ItemBuff", false);
	end
end


-- Checks for buffs on the player
function Highlighting.ButtonMixin:CheckBuff()
	FlexBar2:Debug("checking for buffs");
	local Icon = self:GetIcon();
	if(Icon ~= nil and Aura:UnitHasBuff("player", nil, nil, Icon)) then
		self:SetOuterHighlight("Buff", true);
	else
		-- Check for buffs with the same name as the spell but different texture (hint, weapon buffs :))
		self:SetOuterHighlight("Buff", false);
	end
end


-- Checks for debuffs on the target
function Highlighting.ButtonMixin:CheckDebuff()
	if(self:GetIcon() ~= nil and Aura:UnitHasDebuff("target", nil, nil, self:GetIcon())) then
		self:SetOuterHighlight("Debuff", true);
	else
		self:SetOuterHighlight("Debuff", false);
	end
end

-- Check for equiped items
function Highlighting.ButtonMixin:CheckEquipped()
	local Item = self:GetModifiedAttribute("itemlink");
	if(Item and IsEquippedItem(Item)) then
		self:SetOuterHighlight("Equipped", true);
	else
		self:SetOuterHighlight("Equipped", false);
	end
end

-- Check if the action is in range
function Highlighting.ButtonMixin:CheckRange()
	local ButtonType = self:GetModifiedAttribute("type");
	local IsInRange;
	if(ButtonType == "spell") then
		local Spell = self:GetModifiedAttribute("spell");
		IsInRange = IsSpellInRange(Spell, "target");
	elseif(ButtonType == "item") then
		local Item = self:GetModifiedAttribute("itemlink");
		local IsInRange = IsItemInRange(Item, "target"); 
	else
		self:SetInnerHighlight("NoRange", false);
		return;
	end
	self:SetInnerHighlight("NoRange", IsInRange == 0);
end

-- CHeck if the action is usable 
function Highlighting.ButtonMixin:CheckUsable()
	local ButtonType = self:GetModifiedAttribute("type");
	local Usable, NoMana;
	if(ButtonType == "spell") then
		local Spell = self:GetModifiedAttribute("spell");
		Usable, NoMana = IsUsableSpell(Spell); 
	elseif(ButtonType == "item") then
		local Item = self:GetModifiedAttribute("itemlink");
		Usable, NoMana = IsUsableItem(Item);
	else
		self:SetInnerHighlight("NoMana", false);
		self:SetInnerHighlight("Unusable", false);
		return;
	end
	if(Usable) then
		self:SetInnerHighlight("Unusable", false);
		self:SetInnerHighlight("NoMana", false);
	else
		self:SetInnerHighlight("Unusable", true);
		if(NoMana) then
			self:SetInnerHighlight("NoMana", true);
		else
			self:SetInnerHighlight("NoMana", false);
		end
	end
end
